# -*- coding: utf-8 -*-
"""
 _______  _______  ___      __   __  _______  _______
|       ||       ||   |    |  | |  ||       ||       |
|    ___||    ___||   |    |  | |  ||    _  ||    ___|
|   |___ |   |___ |   |    |  |_|  ||   |_| ||   |___
|    ___||    ___||   |___ |       ||    ___||    ___|
|   |    |   |___ |       ||       ||   |    |   |___
|___|    |_______||_______||_______||___|    |_______|

This file is part of felupe.

Felupe is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Felupe is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with Felupe.  If not, see <http://www.gnu.org/licenses/>.

"""
from copy import deepcopy

import numpy as np
import pytest

import felupe as fem


def pre1d():
    m = fem.mesh.Line()
    e = fem.element.Line()
    q = fem.quadrature.GaussLegendre(1, 1)
    r = fem.Region(m, e, q)
    u = fem.Field(r, dim=1)
    v = fem.FieldContainer([u])
    return v


def pre2d():
    m = fem.mesh.Rectangle()
    e = fem.element.Quad()
    q = fem.quadrature.GaussLegendre(1, 2)
    r = fem.Region(m, e, q)
    u = fem.Field(r, dim=2)
    v = fem.FieldContainer([u])
    return v


def pre3d():
    m = fem.mesh.Cube()
    e = fem.element.Hexahedron()
    q = fem.quadrature.GaussLegendre(1, 3)
    r = fem.Region(m, e, q)
    u = fem.Field(r, dim=2)
    v = fem.FieldContainer([u])
    return v


def test_boundary():
    u = pre3d()
    bounds = {"boundary-label": fem.Boundary(u[0])}

    v = fem.dof.apply(u, bounds, dof0=None)
    assert np.allclose(u[0].values.ravel(), v)

    mask = np.ones(u.region.mesh.npoints, dtype=bool)
    bounds = {"boundary-label": fem.Boundary(u[0], mask=mask)}

    mask = np.ones((u.region.mesh.npoints, u[0].dim), dtype=bool)
    bounds = {"boundary-label": fem.Boundary(u[0], mask=mask)}

    v = fem.dof.apply(u, bounds, dof0=None)
    assert np.allclose(u[0].values.ravel(), v)

    with pytest.raises(ValueError):
        wrong_mask = np.ones((u.region.mesh.npoints, 4), dtype=bool)
        bounds = {"boundary-label": fem.Boundary(u[0], mask=wrong_mask)}


def test_loadcase():
    ux = fem.dof.uniaxial(pre1d(), return_loadcase=False)
    assert len(ux) == 2

    for u in [pre2d(), pre3d()]:
        v = fem.FieldContainer([u[0], deepcopy(u[0])])

        ux = fem.dof.uniaxial(
            u, right=1.0, move=0.2, clamped=False, sym=True, return_loadcase=False
        )
        ux = fem.dof.uniaxial(
            u, right=1.0, move=0.2, clamped=False, sym=True, return_loadcase=True
        )
        assert len(ux) == 2

        ux = fem.dof.uniaxial(
            u, right=1.0, move=0.2, clamped=True, sym=False, return_loadcase=True
        )
        assert len(ux) == 2
        assert "right" in ux[0]

        ux = fem.dof.uniaxial(
            u,
            right=1.0,
            move=0.2,
            clamped=True,
            axis=1,
            sym=(False, True, False),
            return_loadcase=True,
        )
        assert len(ux) == 2
        assert "right" in ux[0]

        ux = fem.dof.uniaxial(
            u,
            right=1.0,
            move=0.2,
            clamped=True,
            axis=1,
            sym=(True, False, True),
            return_loadcase=True,
        )
        assert len(ux) == 2
        assert "right" in ux[0]

        ux = fem.dof.uniaxial(
            u, right=None, move=0.2, clamped=True, return_loadcase=True
        )
        assert len(ux) == 2
        assert "right" in ux[0]

        bx = fem.dof.biaxial(
            u,
            rights=(1.0, 1.0),
            moves=(0.2, 0.2),
            clampes=(False, False),
            return_loadcase=False,
        )
        bx = fem.dof.biaxial(
            u,
            rights=(1.0, 1.0),
            moves=(0.2, 0.2),
            clampes=(False, False),
            return_loadcase=True,
        )
        assert len(bx) == 2

        bx = fem.dof.biaxial(
            u,
            rights=(1.0, 1.0),
            moves=(0.2, 0.2),
            clampes=(True, True),
            sym=False,
            return_loadcase=True,
        )
        assert len(bx) == 2
        assert "left-0" in bx[0]

        bx = fem.dof.biaxial(
            u,
            rights=(None, None),
            moves=(0.2, 0.2),
            clampes=(True, True),
            return_loadcase=True,
        )
        assert len(bx) == 2
        assert "right-0" in bx[0]

        bx = fem.dof.biaxial(
            v,
            rights=(1.0, 1.0),
            moves=(0.2, 0.2),
            clampes=(True, True),
            return_loadcase=True,
        )
        assert len(bx) == 2
        assert "right-0" in bx[0]

        sh = fem.dof.shear(u, bottom=0.0, top=1.0, moves=(0.2, 0, 0), sym=True)
        sh = fem.dof.shear(
            u, bottom=0.0, top=1.0, moves=(0.2, 0, 0), sym=True, return_loadcase=True
        )
        assert len(sh) == 2
        assert "top" in sh[0]

        sh = fem.dof.shear(
            v, bottom=None, top=None, moves=(0.2, 0, 0), sym=False, return_loadcase=True
        )
        assert len(sh) == 2
        assert "top" in sh[0]


def test_boundary_multiaxial():
    mesh = fem.Rectangle(n=3)
    region = fem.RegionQuad(mesh)
    field = fem.FieldContainer([fem.FieldPlaneStrain(region, dim=2)])

    for value in [1.0, np.arange(1, 3), np.arange(1, 3).reshape(1, 2)]:
        boundaries = dict(
            left=fem.Boundary(field[0], fx=0),
            right=fem.Boundary(field[0], fx=1, value=value),
        )

        dof0, dof1 = fem.dof.partition(field, boundaries)
        ext0 = fem.dof.apply(field, boundaries, dof0)

        assert ext0.shape == dof0.shape


def test_boundary_plot():
    region = fem.RegionHexahedron(fem.Cube(b=(3, 1, 1), n=2))
    field = fem.Field(region, dim=3).as_container()
    boundaries = dict(
        left=fem.Boundary(field[0], fx=0, skip=(0, 1, 0)),
        right=fem.Boundary(field[0], fx=1, skip=(0, 0, 0)),
    )
    _ = boundaries["left"].plot(plotter=boundaries["right"].plot())

    field = fem.Field(region, dim=2).as_container()
    boundaries = dict(
        left=fem.Boundary(field[0], fx=0, skip=(0, 1, 0)),
    )

    with pytest.raises(ValueError):
        _ = boundaries["left"].plot()


def test_boundary_dict():
    field = fem.FieldPlaneStrain(
        fem.RegionQuad(fem.Rectangle(b=(3, 1), n=5)), dim=2
    ).as_container()
    boundaries = fem.BoundaryDict(
        left=fem.Boundary(field[0], fx=0, skip=(0, 0)),
        right=fem.Boundary(field[0], fx=3, skip=(0, 1)),
    )
    plotter = boundaries.plot()
    # img = boundaries.screenshot()
    # ax = boundaries.imshow()


def test_symmetry():
    field = fem.FieldPlaneStrain(
        fem.RegionQuad(fem.Rectangle(n=3)), dim=2
    ).as_container()
    boundaries = fem.BoundaryDict()

    with pytest.raises(TypeError):
        fem.dof.symmetry(field)

    with pytest.raises(ValueError):
        fem.dof.symmetry(field[0], bounds=boundaries, boundaries=boundaries)

    with pytest.warns(DeprecationWarning):
        fem.dof.symmetry(field[0], bounds=boundaries)


def test_boundarydict_iter():
    field = fem.FieldPlaneStrain(
        fem.RegionQuad(fem.Rectangle(n=3)), dim=2
    ).as_container()
    with pytest.raises(TypeError):
        boundaries, loadcase = fem.dof.uniaxial(field, return_loadcase=False)

    items = fem.dof.uniaxial(field, return_loadcase=False).items()


def test_loadcase_deprecated():
    field = fem.FieldPlaneStrain(
        fem.RegionQuad(fem.Rectangle(n=3)), dim=2
    ).as_container()

    for loadcase in [fem.dof.uniaxial, fem.dof.shear, fem.dof.biaxial]:

        with pytest.warns(DeprecationWarning):
            my_loadcase = loadcase(field)

        with pytest.warns(DeprecationWarning):
            my_loadcase = loadcase(field, return_loadcase=None)


if __name__ == "__main__":
    test_boundary()
    test_boundary_dict()
    test_boundary_multiaxial()
    test_boundary_plot()
    test_loadcase()
    test_symmetry()
    test_boundarydict_iter()
    test_loadcase_deprecated()



# -*- coding: utf-8 -*-
"""
 _______  _______  ___      __   __  _______  _______
|       ||       ||   |    |  | |  ||       ||       |
|    ___||    ___||   |    |  | |  ||    _  ||    ___|
|   |___ |   |___ |   |    |  |_|  ||   |_| ||   |___
|    ___||    ___||   |___ |       ||    ___||    ___|
|   |    |   |___ |       ||       ||   |    |   |___
|___|    |_______||_______||_______||___|    |_______|

This file is part of felupe.

Felupe is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Felupe is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with Felupe.  If not, see <http://www.gnu.org/licenses/>.

"""
import jax.numpy as jnp
import numpy as np
import pytest

import felupe as fem
import felupe.constitution.jax as mat


def test_vmap():
    def f(x, a=1.0):
        return x

    def g(x, y, a=1.0, **kwargs):
        return x

    vf = fem.constitution.jax.vmap(f)
    vg = fem.constitution.jax.vmap(g)

    x = (np.eye(3).reshape(1, 1, 3, 3) * np.ones((10, 2, 1, 1))).T

    z = vf(x, a=1.0)

    assert np.allclose(z, vf(x, 1.0))
    assert np.allclose(z, vf(a=1.0, x=x))

    with pytest.raises(TypeError):
        vf(x, a=1.0, b=2.0)

    # does not raise an error because of `g(..., **kwargs)`
    assert np.allclose(z, vg(x, a=1.0, b=2.0))


def test_hyperelastic_jax():
    mesh = fem.Cube(n=2)
    region = fem.RegionHexahedron(mesh)
    field = fem.FieldContainer([fem.Field(region, dim=3)])

    md = mat.models.hyperelastic
    for W in [
        md.mooney_rivlin,
        md.yeoh,
        md.third_order_deformation,
        md.miehe_goektepe_lulei,
        md.ogden,
        md.storakers,
        md.van_der_waals,
        md.extended_tube,
        md.blatz_ko,
    ]:
        umat = mat.Hyperelastic(W, **W.kwargs)
        solid = fem.SolidBody(umat=umat, field=field)
        solid.evaluate.gradient()
        solid.evaluate.hessian()

    umat = mat.Hyperelastic(W, **W.kwargs, parallel=True)
    umat = mat.Hyperelastic(W, **W.kwargs, jit=True)


def test_hyperelastic_jax_statevars():
    def W(C, statevars, C10, K):
        I3 = jnp.linalg.det(C)
        J = jnp.sqrt(I3)
        I1 = I3 ** (-1 / 3) * jnp.trace(C)
        statevars_new = statevars.at[0].set(I1)
        return C10 * (I1 - 3) + K * (J - 1) ** 2 / 2, statevars_new

    W.kwargs = {"C10": 0.5}

    umat = mat.Hyperelastic(W, C10=0.5, K=2.0, nstatevars=1, jit=True)
    mesh = fem.Cube(n=2)
    region = fem.RegionHexahedron(mesh)
    field = fem.FieldContainer([fem.Field(region, dim=3)])

    solid = fem.SolidBody(umat=umat, field=field)
    solid.evaluate.gradient()
    solid.evaluate.hessian()


def test_material_jax():
    def dWdF(F, C10, K):
        J = jnp.linalg.det(F)
        C = F.T @ F
        Cu = J ** (-2 / 3) * C
        dev = lambda C: C - jnp.trace(C) / 3 * jnp.eye(3)

        P = 2 * C10 * F @ dev(Cu) @ jnp.linalg.inv(C)
        return P + K * (J - 1) * J * jnp.linalg.inv(C)

    umat = mat.Material(dWdF, C10=0.5, K=2.0, parallel=True)
    umat = mat.Material(dWdF, C10=0.5, K=2.0, jit=True)

    for fun in [dWdF, mat.updated_lagrange(dWdF), mat.total_lagrange(dWdF)]:
        umat = mat.Material(fun, C10=0.5, K=2.0)
        mesh = fem.Cube(n=2)
        region = fem.RegionHexahedron(mesh)
        field = fem.FieldContainer([fem.Field(region, dim=3)])

        solid = fem.SolidBody(umat=umat, field=field)
        solid.evaluate.gradient()
        solid.evaluate.hessian()


def test_material_jax_statevars():
    def dWdF(F, statevars, C10, K):
        J = jnp.linalg.det(F)
        C = F.T @ F
        Cu = J ** (-2 / 3) * C
        dev = lambda C: C - jnp.trace(C) / 3 * jnp.eye(3)

        P = 2 * C10 * F @ dev(Cu) @ jnp.linalg.inv(C)
        statevars_new = statevars.at[0].set(J)
        return P + K * (J - 1) * J * jnp.linalg.inv(C), statevars_new

    dWdF.kwargs = {"C10": 0.5}

    for fun in [dWdF, mat.updated_lagrange(dWdF), mat.total_lagrange(dWdF)]:
        umat = mat.Material(fun, C10=0.5, K=2.0, nstatevars=1, jit=True)
        mesh = fem.Cube(n=2)
        region = fem.RegionHexahedron(mesh)
        field = fem.FieldContainer([fem.Field(region, dim=3)])

        solid = fem.SolidBody(umat=umat, field=field)
        solid.evaluate.gradient()
        solid.evaluate.hessian()


def test_material_included_jax_statevars():
    for fun, nstatevars in zip(
        [
            mat.models.lagrange.morph,
            mat.models.lagrange.morph_representative_directions,
        ],
        [13, 84],
    ):
        umat = mat.Material(
            fun,
            **fun.kwargs,
            nstatevars=nstatevars,
        )
        mesh = fem.Cube(n=2)
        region = fem.RegionHexahedron(mesh)
        field = fem.FieldContainer([fem.Field(region, dim=3)])

        solid = fem.SolidBody(umat=umat, field=field)
        solid.evaluate.gradient()
        solid.evaluate.hessian()


if __name__ == "__main__":
    test_vmap()
    test_hyperelastic_jax()
    test_hyperelastic_jax_statevars()
    test_material_jax()
    test_material_jax_statevars()
    test_material_included_jax_statevars()



# -*- coding: utf-8 -*-
"""
This file is part of FElupe.

FElupe is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

FElupe is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with FElupe.  If not, see <http://www.gnu.org/licenses/>.
"""

from ..math import displacement
from ._mesh import ViewMesh


class ViewField(ViewMesh):
    r"""Provide Visualization methods for :class:`felupe.FieldContainer`. The warped
    (deformed) mesh is created from the values of the first field (displacements). By
    default, the "Deformation Gradient" tensor, the "Logarithmic Strain" tensor and the
    "Principal Values of Logarithmic Strain" are evaluated as field-related items of the
    cell-data dict. Optional items of given point- and cell-data overwrite these default
    field-related cell-data items.

    Parameters
    ----------
    field : felupe.FieldContainer
        The field-container.
    point_data : dict or None, optional
        Additional point-data dict (default is None).
    cell_data : dict or None, optional
        Additional cell-data dict (default is None).
    cell_type : pyvista.CellType or None, optional
        Cell-type of PyVista (default is None).
    project : callable or None, optional
        Callable to project internal cell-data at quadrature-points to mesh-points
        (default is None). Valid callables are :class:`~felupe.project` or
        :class:`~felupe.tools.extrapolate`.

    Attributes
    ----------
    mesh : pyvista.UnstructuredGrid
        A generalized Dataset with the mesh as well as point- and cell-data. This is
        not an instance of :class:`felupe.Mesh`.

    Examples
    --------
    ..  pyvista-plot::
        :force_static:

        >>> import numpy as np
        >>> import felupe as fem
        >>>
        >>> mesh = fem.Cube(n=3)
        >>> region = fem.RegionHexahedron(mesh)
        >>> u = np.sqrt(1 + np.arange(81)).reshape(27, 3) / 100
        >>> field = fem.FieldContainer([fem.Field(region, values=u)])
        >>>
        >>> view = fem.ViewField(field, project=fem.project)
        >>> view.plot("Principal Values of Logarithmic Strain").show()

    See Also
    --------
    felupe.view.Scene : Base class for plotting a static scene.
    felupe.ViewMesh : Provide Visualization methods for a mesh with optional given
        dicts of point- and cell-data items.
    felupe.ViewSolid : Provide Visualization methods for a field container or a
        solid body.
    felupe.project: Project given values at quadrature-points to mesh-points.
    """

    def __init__(
        self, field, point_data=None, cell_data=None, cell_type=None, project=None
    ):
        point_data_from_field = {}
        cell_data_from_field = {}

        if hasattr(field.region, "dhdX"):

            if project is None:
                cell_data_from_field = {
                    "Deformation Gradient": field.evaluate.deformation_gradient()
                    .mean(-2)
                    .T,
                    "Logarithmic Strain": field.evaluate.strain(
                        tensor=True, asvoigt=True
                    )
                    .mean(-2)
                    .T,
                    "Principal Values of Logarithmic Strain": field.evaluate.strain(
                        tensor=False
                    )
                    .mean(-2)
                    .T,
                }
            elif callable(project):
                point_data_from_field = {
                    "Deformation Gradient": project(
                        field.evaluate.deformation_gradient(), field.region
                    ),
                    "Logarithmic Strain": project(
                        field.evaluate.strain(tensor=True, asvoigt=True), field.region
                    ),
                    "Principal Values of Logarithmic Strain": project(
                        field.evaluate.strain(tensor=False), field.region
                    ),
                }
            else:
                raise TypeError("The project-argument must be callable or None.")

        point_data_from_field["Displacement"] = displacement(field)

        if point_data is None:
            point_data = {}

        if cell_data is None:
            cell_data = {}

        super().__init__(
            mesh=field.region.mesh,
            point_data={**point_data_from_field, **point_data},
            cell_data={**cell_data_from_field, **cell_data},
            cell_type=cell_type,
        )



from tensortrax.math import einsum, sqrt, trace
from tensortrax.math.linalg import det, inv

from ......quadrature import BazantOh
from ...._total_lagrange import total_lagrange


@total_lagrange
def affine_force_statevars(F, statevars, f, kwargs, quadrature=BazantOh(n=21)):
    "Micro-sphere model: Affine force (stretch) part."

    r = quadrature.points
    M = einsum("ai,aj->aij", r, r)
    Mw = einsum("aij,a->aij", M, quadrature.weights)

    # affine stretches (unimodular part)
    J = det(F)
    C = F.T @ F
    λ = J ** (-1 / 3) * sqrt(einsum("ij...,aij->a...", C, M))

    dψdλ, statevars_new = f(λ, statevars, **kwargs)
    dψdE = einsum("a...,aij->ij...", dψdλ / λ, Mw)

    S = J ** (-2 / 3) * (dψdE - trace(dψdE @ C) / 3 * inv(C))

    return S, statevars_new



# -*- coding: utf-8 -*-
"""
This file is part of FElupe.

FElupe is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

FElupe is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with FElupe.  If not, see <http://www.gnu.org/licenses/>.
"""

from tensortrax.math import sum as tsum
from tensortrax.math import trace
from tensortrax.math.linalg import det


def arruda_boyce(C, C1, limit):
    r"""Strain energy function of the isotropic hyperelastic
    `Arruda-Boyce <https://en.wikipedia.org/wiki/Arruda-Boyce_model>`_ material
    formulation.

    Parameters
    ----------
    C : tensortrax.Tensor
        Right Cauchy-Green deformation tensor.
    C1 : float
        Initial shear modulus.
    limit : float
        Limiting stretch :math:`\lambda_m` at which the polymer chain network becomes
        locked.

    Notes
    -----
    The strain energy function is given in Eq. :eq:`psi-ab`
    
    ..  math::
        :label: psi-ab

        \psi = C_1 \sum_{i=1}^5 \alpha_i \beta^{i-1} \left( \hat{I}_1^i - 3^i \right)

    with the first main invariant of the distortional part of the right
    Cauchy-Green deformation tensor as given in Eq. :eq:`invariants-ab`
    
    ..  math::
        :label: invariants-ab
    
        \hat{I}_1 = J^{-2/3} \text{tr}\left( \boldsymbol{C} \right)
    
    and :math:`\alpha_i` and :math:`\beta` as denoted in Eq. :eq:`ab-param`.
    
    ..  math::
        :label: ab-param
        
        \boldsymbol{\alpha} &= \begin{bmatrix} 
            \frac{1}{2} \\ 
            \frac{1}{20} \\
            \frac{11}{1050} \\
            \frac{19}{7000} \\
            \frac{519}{673750}
        \end{bmatrix}
        
        \beta &= \frac{1}{\lambda_m^2}
    
    The initial shear modulus is a function of both material parameters, see Eq.
    :eq:`shear-modulus-ab`.
    
    ..  math::
        :label: shear-modulus-ab
        
        \mu = C_1 \left( 
            1 + \frac{3}{5 \lambda_m^2} + \frac{99}{175 \lambda_m^4} 
              + \frac{513}{875 \lambda_m^6} + \frac{42039}{67375 \lambda_m^8} 
        \right)

    Examples
    --------
    
    ..  plot::

        >>> import felupe as fem
        >>>
        >>> umat = fem.Hyperelastic(fem.arruda_boyce, C1=1.0, limit=3.2)
        >>> ax = umat.plot(incompressible=True)

    """
    I1 = det(C) ** (-1 / 3) * trace(C)

    alphas = [1 / 2, 1 / 20, 11 / 1050, 19 / 7000, 519 / 673750]
    beta = 1 / limit**2

    out = []
    for j, alpha in enumerate(alphas):
        i = j + 1
        out.append(alpha * beta ** (i - 1) * (I1**i - 3**i))

    return C1 * tsum(out)



# -*- coding: utf-8 -*-
"""
This file is part of FElupe.

FElupe is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

FElupe is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with FElupe.  If not, see <http://www.gnu.org/licenses/>.
"""

import numpy as np


class Element:
    """
    Base-class for a finite element which provides methods for plotting.

    Examples
    --------
    This example shows how to implement a hexahedron element.

    ..  pyvista-plot::
        :force_static:

        >>> import felupe as fem
        >>> import numpy as np
        >>>
        >>> class Hexahedron(fem.Element):
        ...     def __init__(self):
        ...         a = [-1, 1, 1, -1, -1, 1, 1, -1]
        ...         b = [-1, -1, 1, 1, -1, -1, 1, 1]
        ...         c = [-1, -1, -1, -1, 1, 1, 1, 1]
        ...         self.points = np.vstack([a, b, c]).T
        ...
        ...         # additional attributes for plotting, optional
        ...         self.cells = np.array([[0, 1, 2, 3, 4, 5, 6, 7]])
        ...         self.cell_type = "hexahedron"
        ...
        ...     def function(self, rst):
        ...         r, s, t = rst
        ...         a, b, c = self.points.T
        ...         ar, bs, ct = 1 + a * r, 1 + b * s, 1 + c * t
        ...         return (ar * bs * ct) / 8
        ...
        ...     def gradient(self, rst):
        ...         r, s, t = rst
        ...         a, b, c = self.points.T
        ...         ar, bs, ct = 1 + a * r, 1 + b * s, 1 + c * t
        ...         return np.stack([a * bs * ct, ar * b * ct, ar * bs * c], axis=1)
        >>>
        >>> mesh = fem.Cube(n=6)
        >>> element = Hexahedron()
        >>> quadrature = fem.GaussLegendre(order=1, dim=3)
        >>> region = fem.Region(mesh, element, quadrature)

    The gradient (and optionally the hessian) may be carried out using automatic
    differentiation.

    ..  warning::
        It is important to use only differentiable math-functions from
        :mod:`tensortrax.math`!

    ..  pyvista-plot::
        :force_static:

        >>> import felupe as fem
        >>> import tensortrax as tr
        >>> import numpy as np
        >>>
        >>> class Hexahedron(fem.Element):
        ...     def __init__(self):
        ...         a = [-1, 1, 1, -1, -1, 1, 1, -1]
        ...         b = [-1, -1, 1, 1, -1, -1, 1, 1]
        ...         c = [-1, -1, -1, -1, 1, 1, 1, 1]
        ...         self.points = np.vstack([a, b, c]).T
        ...
        ...         # additional attributes for plotting, optional
        ...         self.cells = np.array([[0, 1, 2, 3, 4, 5, 6, 7]])
        ...         self.cell_type = "hexahedron"
        ...
        ...     def fun(self, rst):
        ...         a, b, c = tr.math.array(self.points.T, shape=(3, 8), like=rst)
        ...         r, s, t = rst
        ...         ar, bs, ct = 1 + a * r, 1 + b * s, 1 + c * t
        ...         return (ar * bs * ct) / 8
        ...
        ...     def function(self, rst):
        ...         return tr.function(self.fun)(rst)
        ...
        ...     def gradient(self, rst):
        ...         return tr.jacobian(self.fun)(rst)
        ...
        ...     def hessian(self, rst):
        ...         return tr.hessian(self.fun)(rst)
        >>>
        >>> mesh = fem.Cube(n=6)
        >>> element = Hexahedron()
        >>> quadrature = fem.GaussLegendre(order=1, dim=3)
        >>> region = fem.Region(mesh, element, quadrature, hess=True)

    """

    def view(self, point_data=None, cell_data=None, cell_type=None):
        """View the element with optional given dicts of point- and cell-data items.

        Parameters
        ----------
        point_data : dict or None, optional
            Additional point-data dict (default is None).
        cell_data : dict or None, optional
            Additional cell-data dict (default is None).
        cell_type : pyvista.CellType or None, optional
            Cell-type of PyVista (default is None).

        Returns
        -------
        felupe.ViewMesh
            A object which provides visualization methods for
            :class:`felupe.element.Element`.

        See Also
        --------
        felupe.ViewMesh : Visualization methods for :class:`felupe.Mesh`.
        """

        from ..mesh import Mesh

        mesh = Mesh(points=self.points, cells=self.cells, cell_type=self.cell_type)

        return mesh.view(
            point_data=point_data,
            cell_data=cell_data,
            cell_type=cell_type,
        )

    def plot(
        self,
        *args,
        style="wireframe",
        color="black",
        add_axes_at_origin=True,
        add_point_labels=True,
        show_points=True,
        font_size=26,
        **kwargs,
    ):
        """Plot the element.

        See Also
        --------
        felupe.Scene.plot: Plot method of a scene.
        """

        view = self.view()
        if self.points.shape[1] > 1:
            view.mesh = view.mesh.extract_surface().extract_feature_edges()

        plotter = view.plot(
            *args,
            show_undeformed=False,
            opacity=0.8,
            add_axes=False,
            show_edges=False,
            style=style,
            color=color,
            **kwargs,
        )
        if add_point_labels:
            plotter.add_point_labels(
                points=np.pad(self.points, ((0, 0), (0, 3 - self.points.shape[1]))),
                labels=[f"{a}" for a in np.arange(len(self.points))],
                font_size=font_size,
                show_points=show_points,
                point_size=20,
                point_color="black",
                shape=None,
                fill_shape=False,
                render_points_as_spheres=True,
                always_visible=True,
            )

        if add_axes_at_origin:
            actor = plotter.add_axes_at_origin(xlabel="r", ylabel="s", zlabel="t")
            actor.SetNormalizedShaftLength((0.9, 0.9, 0.9))
            actor.SetNormalizedTipLength((0.1, 0.1, 0.1))

            if self.points.shape[1] == 3:
                actor.SetTotalLength([1.3, 1.3, 1.3])
            elif self.points.shape[1] == 2:
                actor.SetZAxisLabelText("")
                actor.SetTotalLength([1.3, 1.3, 0])
            elif self.points.shape[1] == 1:
                actor.SetYAxisLabelText("")
                actor.SetZAxisLabelText("")
                actor.SetTotalLength([1.3, 0, 0])

            plotter.camera.zoom(0.7)

        if self.points.shape[1] == 3:
            plotter.camera.azimuth = -17

        return plotter

    def screenshot(
        self,
        *args,
        filename=None,
        transparent_background=None,
        scale=None,
        **kwargs,
    ):
        """Take a screenshot of the element.

        See Also
        --------
        pyvista.Plotter.screenshot: Take a screenshot of a PyVista plotter.
        """

        if filename is None:
            filename = f"{self.cell_type}.png"

        return self.plot(*args, off_screen=True, **kwargs).screenshot(
            filename=filename,
            transparent_background=transparent_background,
            scale=scale,
        )



from tensortrax.math import einsum, sqrt
from tensortrax.math.linalg import det, inv

from ......quadrature import BazantOh


def affine_stretch(C, f, kwargs, quadrature=BazantOh(n=21)):
    "Micro-sphere model: Affine stretch part."

    r = quadrature.points
    w = quadrature.weights

    # affine stretches (distortional part)
    λ = det(C) ** (-1 / 6) * sqrt(einsum("ai,ij...,aj->a...", r, C, r))

    return einsum("a...,a->...", f(λ, **kwargs), w)


def affine_stretch_statevars(C, statevars, f, kwargs, quadrature=BazantOh(n=21)):
    "Micro-sphere model: Affine stretch part."

    r = quadrature.points
    w = quadrature.weights

    # affine stretches (distortional part)
    λ = det(C) ** (-1 / 6) * sqrt(einsum("ai,ij...,aj->a...", r, C, r))
    ψ, statevars_new = f(λ, statevars, **kwargs)

    return einsum("a...,a->...", ψ, w), statevars_new


def affine_tube(C, f, kwargs, quadrature=BazantOh(n=21)):
    "Micro-sphere model: Affine area-stretch part."

    r = quadrature.points
    w = quadrature.weights

    # affine area-stretches (distortional part)
    λa = det(C) ** (1 / 6) * sqrt(einsum("ai,ij...,aj->a...", r, inv(C), r))

    return einsum("a...,a->...", f(λa, **kwargs), w)


def affine_tube_statevars(C, statevars, f, kwargs, quadrature=BazantOh(n=21)):
    "Micro-sphere model: Affine area-stretch part."

    r = quadrature.points
    w = quadrature.weights

    # affine area-stretches (distortional part)
    λa = det(C) ** (1 / 6) * sqrt(einsum("ai,ij...,aj->a...", r, inv(C), r))
    ψa, statevars_new = f(λa, statevars, **kwargs)

    return einsum("a...,a->...", ψa, w), statevars_new



# -*- coding: utf-8 -*-
"""
This file is part of FElupe.

FElupe is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

FElupe is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with FElupe.  If not, see <http://www.gnu.org/licenses/>.
"""
from tensortrax.math import array, maximum, sqrt
from tensortrax.math.linalg import det, eigvalsh, expm, inv
from tensortrax.math.special import dev, from_triu_1d, sym, triu_1d, try_stack

from ..._total_lagrange import total_lagrange


@total_lagrange
def morph(F, statevars, p):
    r"""Second Piola-Kirchhoff stress tensor of the
    `MORPH <https://doi.org/10.1016/s0749-6419(02)00091-8>`_ model formulation [1]_.

    Parameters
    ----------
    F : tensortrax.Tensor or jax.Array
        Deformation gradient tensor.
    statevars : array
        Vector of stacked state variables (CTS, C, SA).
    p : list of float
        A list which contains the 8 material parameters.

    Notes
    -----
    The MORPH material model is implemented as a second Piola-Kirchhoff stress-based
    formulation with automatic differentiation. The Tresca invariant of the distortional
    part of the right Cauchy-Green deformation tensor is used as internal state
    variable, see Eq. :eq:`morph-state`.

    ..  warning::
        While the `MORPH <https://doi.org/10.1016/s0749-6419(02)00091-8>`_-material
        formulation captures the Mullins effect and quasi-static hysteresis effects of
        rubber mixtures very nicely, it has been observed to be unstable for medium- to
        highly-distorted states of deformation.

    ..  math::
        :label: morph-state

        \boldsymbol{C} &= \boldsymbol{F}^T \boldsymbol{F}

        I_3 &= \det (\boldsymbol{C})

        \hat{\boldsymbol{C}} &= I_3^{-1/3} \boldsymbol{C}

        \hat{\lambda}^2_\alpha &= \text{eigvals}(\hat{\boldsymbol{C}})

        \hat{C}_T &= \max \left( \hat{\lambda}^2_\alpha - \hat{\lambda}^2_\beta \right)

        \hat{C}_T^S &= \max \left( \hat{C}_T, \hat{C}_{T,n}^S \right)

    A sigmoid-function is used inside the deformation-dependent variables
    :math:`\alpha`, :math:`\beta` and :math:`\gamma`, see Eq. :eq:`morph-sigmoid`.

    ..  math::
        :label: morph-sigmoid

        f(x) &= \frac{1}{\sqrt{1 + x^2}}

        \alpha &= p_1 + p_2 \ f(p_3\ C_T^S)

        \beta &= p_4\ f(p_3\ C_T^S)

        \gamma &= p_5\ C_T^S\ \left( 1 - f\left(\frac{C_T^S}{p_6}\right) \right)

    The rate of deformation is described by the Lagrangian tensor and its Tresca-
    invariant, see Eq. :eq:`morph-rate-of-deformation`.

    ..  note::
        It is important to evaluate the incremental right Cauchy-Green tensor by the
        difference of the final and the previous state of deformation, not by its
        variation with respect to the deformation gradient tensor.

    ..  math::
        :label: morph-rate-of-deformation

        \hat{\boldsymbol{L}} &= \text{sym}\left(
                \text{dev}(\boldsymbol{C}^{-1} \Delta\boldsymbol{C})
            \right) \hat{\boldsymbol{C}}

        \lambda_{\hat{\boldsymbol{L}}, \alpha} &= \text{eigvals}(\hat{\boldsymbol{L}})

        \hat{L}_T &= \max \left(
            \lambda_{\hat{\boldsymbol{L}}, \alpha}-\lambda_{\hat{\boldsymbol{L}}, \beta}
        \right)

        \Delta\boldsymbol{C} &= \boldsymbol{C} - \boldsymbol{C}_n

    The additional stresses evolve between the limiting stresses, see Eq.
    :eq:`morph-stresses`. The additional deviatoric-enforcement terms [1]_ are neglected
    in this implementation.

    ..  math::
        :label: morph-stresses

        \boldsymbol{S}_L &= \left(
            \gamma \exp \left(p_7 \frac{\hat{\boldsymbol{L}}}{\hat{L}_T}
                \frac{\hat{C}_T}{\hat{C}_T^S} \right) +
                p8 \frac{\hat{\boldsymbol{L}}}{\hat{L}_T}
        \right) \boldsymbol{C}^{-1}

        \boldsymbol{S}_A &= \frac{
            \boldsymbol{S}_{A,n} + \beta\ \hat{L}_T\ \boldsymbol{S}_L
        }{1 + \beta\ \hat{L}_T}

        \boldsymbol{S} &= 2 \alpha\ \text{dev}( \hat{\boldsymbol{C}} )
            \boldsymbol{C}^{-1}+\text{dev}\left(\boldsymbol{S}_A\ \boldsymbol{C}\right)
            \boldsymbol{C}^{-1}

    ..  note::
        Only the upper-triangle entries of the symmetric stress-tensor state
        variables are stored in the solid body. Hence, it is necessary to extract such
        variables with :func:`tm.special.from_triu_1d` and export them as
        :func:`tm.special.triu_1d`.

    Examples
    --------
    First, choose the desired automatic differentiation backend

    ..  plot::
        :context: close-figs

        >>> import felupe as fem
        >>>
        >>> # import felupe.constitution.jax as mat
        >>> import felupe.constitution.tensortrax as mat

    and create the material.

    ..  plot::
        :context: close-figs

        >>> umat = mat.Material(
        ...     mat.models.lagrange.morph,
        ...     p=[0.039, 0.371, 0.174, 2.41, 0.0094, 6.84, 5.65, 0.244],
        ...     nstatevars=13,
        ... )
        >>> ax = umat.plot(
        ...    incompressible=True,
        ...    ux=fem.math.linsteps(
        ...        # [1, 2, 1, 2.75, 1, 3.5, 1, 4.2, 1, 4.8, 1, 4.8, 1],
        ...        [1, 2.75, 1, 2.75],
        ...        num=20,
        ...    ),
        ...    ps=None,
        ...    bx=None,
        ... )

    References
    ----------
    .. [1] D. Besdo and J. Ihlemann, "A phenomenological constitutive model for
       rubberlike materials and its numerical applications", International Journal
       of Plasticity, vol. 19, no. 7. Elsevier BV, pp. 1019–1036, Jul. 2003. doi:
       `10.1016/s0749-6419(02)00091-8 <https://doi.org/10.1016/s0749-6419(02)00091-8>`_.

    See Also
    --------
    felupe.constitution.tensortrax.models.lagrange.morph_representative_directions :
        Strain energy function of the MORPH model formulation, implemented by the
        concept of representative directions.
    felupe.constitution.jax.models.lagrange.morph_representative_directions : Strain
        energy function of the MORPH model formulation, implemented by the concept of
        representative directions.
    """

    # right Cauchy-Green deformation tensor
    C = F.T @ F

    # extract old state variables
    CTSn = array(statevars[0], like=C[0, 0])
    Cn = from_triu_1d(statevars[1:7], like=C)
    SAn = from_triu_1d(statevars[7:13], like=C)

    # distortional part of right Cauchy-Green deformation tensor
    I3 = det(C)
    CG = C * I3 ** (-1 / 3)

    # inverse of and incremental right Cauchy-Green deformation tensor
    invC = inv(C)
    dC = C - Cn

    # eigenvalues of right Cauchy-Green deformation tensor (sorted in ascending order)
    λCG = eigvalsh(CG)

    # Tresca invariant of distortional part of right Cauchy-Green deformation tensor
    CTG = λCG[-1] - λCG[0]

    # maximum Tresca invariant in load history
    CTS = maximum(CTG, CTSn)

    def sigmoid(x):
        "Algebraic sigmoid function."
        return 1 / sqrt(1 + x**2)

    # material parameters
    α = p[0] + p[1] * sigmoid(p[2] * CTS)
    β = p[3] * sigmoid(p[2] * CTS)
    γ = p[4] * CTS * (1 - sigmoid(CTS / p[5]))

    LG = sym(dev(invC @ dC)) @ CG
    λLG = eigvalsh(LG)
    LTG = λLG[-1] - λLG[0]

    # limiting stresses "L" and additional stresses "A"
    SL = (γ * expm(p[6] * LG / LTG * CTG / CTS) + p[7] * LG / LTG) @ invC
    SA = (SAn + β * LTG * SL) / (1 + β * LTG)

    # second Piola-Kirchhoff stress tensor
    S = 2 * α * dev(CG) @ invC + dev(SA @ C) @ invC
    statevars_new = try_stack([[CTS], triu_1d(C), triu_1d(SA)], fallback=statevars)

    return S, statevars_new



from ._lame_converter import lame_converter, lame_converter_orthotropic
from ._linear_elastic import (
    LinearElastic,
    LinearElasticPlaneStrain,
    LinearElasticPlaneStress,
    LinearElasticTensorNotation,
)
from ._linear_elastic_1d import LinearElastic1D
from ._linear_elastic_large_strain import LinearElasticLargeStrain
from ._linear_elastic_orthotropic import LinearElasticOrthotropic

__all__ = [
    "lame_converter",
    "lame_converter_orthotropic",
    "LinearElastic",
    "LinearElastic1D",
    "LinearElasticLargeStrain",
    "LinearElasticOrthotropic",
    "LinearElasticPlaneStrain",
    "LinearElasticPlaneStress",
    "LinearElasticTensorNotation",
]



import warnings


class BoundaryDict(dict):
    "A dict of boundary conditions."

    def __iter__(self):
        raise TypeError(
            "Iteration over BoundaryDict is not allowed, use `.items()` instead. "
            "Did you try to run"
            "`boundaries, loadcase = fem.dof.uniaxial(field, return_loadcase=False)`? "
            "Starting with v11.0.0, the return of `loadcase` in `fem.dof.uniaxial()`, "
            "`fem.dof.shear()` and `fem.dof.biaxial()` is optional and must be "
            "explicitely activated by `return_loadcase=True`."
        )

    def plot(
        self,
        plotter=None,
        colors=None,
        size=(0.1, 0.1),
        show_points=True,
        show_lines=True,
        **kwargs,
    ):
        "Plot the boundary conditions."

        if colors is None:
            import matplotlib.colors as mcolors

            colors = list(mcolors.TABLEAU_COLORS.values())

        colors_list = []
        while len(colors_list) < len(self.keys()):
            colors_list = [*colors_list, *colors]

        for (label, boundary), color in zip(self.items(), colors_list):

            plotter = boundary.plot(
                label=label,
                color=color,
                plotter=plotter,
                size=size,
                show_points=show_points,
                show_lines=show_lines,
                **kwargs,
            )

        plotter.add_legend(size=size, bcolor="white")

        return plotter

    def screenshot(
        self,
        *args,
        filename="boundaries.png",
        transparent_background=None,
        scale=None,
        colors=None,
        plotter=None,
        **kwargs,
    ):
        "Take a screenshot of the boundary conditions."

        if plotter is None:
            mesh = self[list(self.keys())[0]].field.region.mesh
            plotter = mesh.plot(off_screen=True)

        plotter = self.plot(plotter=plotter, colors=colors, **kwargs)

        return plotter.screenshot(
            filename=filename,
            transparent_background=transparent_background,
            scale=scale,
        )

    def imshow(self, *args, ax=None, dpi=None, **kwargs):
        """Take a screenshot of the boundary conditions, show the image data in
        a figure and return the ax.
        """

        if ax is None:
            import matplotlib.pyplot as plt

            fig, ax = plt.subplots(dpi=dpi)

        ax.imshow(self.screenshot(*args, filename=None, **kwargs))
        ax.set_axis_off()

        return ax



# -*- coding: utf-8 -*-
"""
This file is part of FElupe.

FElupe is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

FElupe is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with FElupe.  If not, see <http://www.gnu.org/licenses/>.
"""

from functools import wraps

from tensortrax.math.linalg import det


def isochoric_volumetric_split(fun):
    """Apply the material formulation only on the isochoric part of the
    multiplicative split of the deformation gradient."""

    @wraps(fun)
    def apply_iso(C, *args, **kwargs):
        return fun(det(C) ** (-1 / 3) * C, *args, **kwargs)

    return apply_iso
